package org.warp.commonutils.concurrency.executor;

import java.lang.StackWalker.StackFrame;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import org.jetbrains.annotations.NotNull;
import org.warp.commonutils.type.IntWrapper;

public class AsyncStackTraceExecutorDecorator extends ExecutorDecorator {

	private static final ReentrantReadWriteLock lock = new ReentrantReadWriteLock();
	private static final Map<Thread, LSTTask> threadToTask = new HashMap<>();
	private static final Map<Thread, AsyncStackTraceExecutorDecorator> threadToExecutor = new HashMap<>();

	public AsyncStackTraceExecutorDecorator(Executor executor) {
		super(executor);
	}

	@Override
	public void execute(@NotNull Runnable command) {
		var currentThread = Thread.currentThread();
		List<StackFrame> frames = new ArrayList<>();

		LSTTask lstTask;

		lock.readLock().lock();
		try {
			lstTask = threadToTask.getOrDefault(currentThread, null);

			// Add the current stack frames
			addCurrentStackTrace(frames, lstTask != null);

			if (lstTask != null) {
				frames.addAll(lstTask.frames);
			}
			lstTask = new LSTTask(command, frames);

			//System.out.println("execute(): THREAD-" + Thread.currentThread().hashCode() + " TASK-" + lstTask.hashCode());
		} finally {
			lock.readLock().unlock();
		}

		super.execute(lstTask);
	}

	private static void addCurrentStackTrace(List<StackFrame> frames, boolean isFromAsyncCal) {
		StackWalker.getInstance().walk(a -> {
			IntWrapper count = new IntWrapper(0);
			int STACK_MAX_SIZE = 10;
			a.filter(x -> {
				var cn = x.getClassName();
				return !cn.equals("java.util.concurrent.CompletableFuture")
						&& !cn.equals("java.util.concurrent.CompletableFuture$AsyncRun")
						&& !cn.equals("java.util.concurrent.ThreadPoolExecutor")
						&& !cn.equals("java.util.concurrent.ThreadPoolExecutor$Worker")
						&& !cn.equals("java.lang.Thread")
						&& !cn.equals(LSTTask.class.getName())
						&& !cn.equals(AsyncStackTraceExecutorDecorator.class.getName());
			}).skip(0).limit(STACK_MAX_SIZE + 1).peek(x -> count.var++).forEachOrdered(frames::add);
			if (count.var > STACK_MAX_SIZE) {
				frames.remove(frames.size() - 1);
				frames.add(new TextStackFrame("AndMoreFrames"));
			}
			return null;
		});
		if (isFromAsyncCal) {
			frames.add(new TextStackFrame("AsyncCall"));
		}
	}


	class LSTTask implements Runnable {
		private final Runnable runnable;
		List<StackFrame> frames;

		LSTTask(Runnable runnable, List<StackFrame> frames) {
			this.runnable = runnable;
			this.frames = frames;
		}

		@Override
		public void run() {
			var currentThread = Thread.currentThread();

			lock.writeLock().lock();
			try {
				threadToTask.put(currentThread, LSTTask.this);
				threadToExecutor.put(currentThread, AsyncStackTraceExecutorDecorator.this);
			} finally {
				lock.writeLock().unlock();
			}
			try {
				//System.out.println("    run(): THREAD-" + Thread.currentThread().hashCode() + " TASK-" + this.hashCode());
				runnable.run();
			} catch (Throwable t) {
				RuntimeException e = new RuntimeException(t);
				e.setStackTrace(frames.stream().map(StackFrame::toStackTraceElement).toArray(StackTraceElement[]::new));
				throw e;
			}
			lock.writeLock().lock();
			try {
				threadToExecutor.remove(currentThread, AsyncStackTraceExecutorDecorator.this);
				threadToTask.remove(currentThread, LSTTask.this);
			} finally {
				lock.writeLock().unlock();
			}
		}
	}

	public static void fixStackTrace(Exception ex) {
		List<StackTraceElement> result = new ArrayList<>();
		var currentThread = Thread.currentThread();

		lock.readLock().lock();
		try {
			var executor = threadToExecutor.getOrDefault(currentThread, null);
			if (executor != null) {
				LSTTask lstTask = threadToTask.getOrDefault(currentThread, null);
				if (lstTask != null) {
					var currentStackFrames = new ArrayList<StackFrame>();
					addCurrentStackTrace(currentStackFrames, true);
					for (var frame : currentStackFrames) {
						result.add(frame.toStackTraceElement());
					}

					for (var frame : lstTask.frames) {
						result.add(frame.toStackTraceElement());
					}
					ex.setStackTrace(result.toArray(StackTraceElement[]::new));
				}
			}
		} finally {
			lock.readLock().unlock();
		}
	}

	public static void dumpStack() {
		var currentThread = Thread.currentThread();

		lock.readLock().lock();
		try {
			var executor = threadToExecutor.getOrDefault(currentThread, null);
			if (executor != null) {
				LSTTask lstTask = threadToTask.getOrDefault(currentThread, null);
				if (lstTask != null) {
					StringBuilder sb = new StringBuilder();
					sb.append(new Exception("Stack trace").toString()).append('\n');
					var currentStackFrames = new ArrayList<StackFrame>();
					addCurrentStackTrace(currentStackFrames, true);
					for (var frame : currentStackFrames) {
						printStackFrame(sb, frame);
					}

					for (var frame : lstTask.frames) {
						printStackFrame(sb, frame);
					}
					System.err.println(sb.toString());
					return;
				}
			}
			Thread.dumpStack();
		} finally {
			lock.readLock().unlock();
		}
	}

	private static void printStackFrame(StringBuilder sb, StackFrame frame) {
		if(frame.getClassName().equals("AsyncCall")) {
			sb.append("\t(async call)\n");
		} else if(frame.getClassName().equals("AndMoreFrames")) {
			sb.append("\t... omitted more frames\n");
		} else {
			sb.append("\tat ").append(frame.toString()).append('\n');
		}
	}

	private static class TextStackFrame implements StackFrame {

		private final String text;

		public TextStackFrame(String text) {
			this.text = text;
		}

		@Override
		public String getClassName() {
			return text;
		}

		@Override
		public String getMethodName() {
			return "..";
		}

		@Override
		public Class<?> getDeclaringClass() {
			return Object.class;
		}

		@Override
		public int getByteCodeIndex() {
			return 0;
		}

		@Override
		public String getFileName() {
			return null;
		}

		@Override
		public int getLineNumber() {
			return 0;
		}

		@Override
		public boolean isNativeMethod() {
			return false;
		}

		@Override
		public StackTraceElement toStackTraceElement() {
			return new StackTraceElement(getClassName(), getMethodName(), getFileName(), getLineNumber());
		}
	}
}