package org.warp.commonutils.type;

import com.google.common.collect.Queues;
import java.lang.reflect.Array;
import java.util.Collection;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.jetbrains.annotations.NotNull;

public class FloatPriorityQueue<T> implements Queue<T> {

	private final Queue<ScoredValue<T>> internalQueue;

	public static <T> FloatPriorityQueue<T> of() {
		return new FloatPriorityQueue<T>(0);
	}

	public static <T> FloatPriorityQueue<T> of(T value, float score) {
		var pq = new FloatPriorityQueue<T>(1);
		pq.offer(value);
		return pq;
	}

	public static <T> FloatPriorityQueue<T> of(ScoredValue<T> value) {
		var pq = new FloatPriorityQueue<T>(1);
		pq.offer(value);
		return pq;
	}

	public static <T> FloatPriorityQueue<T> of(ScoredValue<T>... values) {
		var pq = new FloatPriorityQueue<T>(values.length);
		for (ScoredValue<T> value : values) {
			pq.offer(value);
		}
		return pq;
	}

	public FloatPriorityQueue(PriorityQueue<ScoredValue<T>> internalQueue) {
		this.internalQueue = internalQueue;
	}

	private FloatPriorityQueue(Queue<ScoredValue<T>> internalQueue) {
		this.internalQueue = internalQueue;
	}

	public FloatPriorityQueue() {
		internalQueue = new PriorityQueue<>();
	}

	public FloatPriorityQueue(int initialCapacity) {
		internalQueue = new PriorityQueue<>(Math.max(1, initialCapacity));
	}

	public static <T> FloatPriorityQueue<T> synchronize(FloatPriorityQueue<T> queue) {
		return new FloatPriorityQueue<T>(Queues.synchronizedQueue(queue.internalQueue));
	}

	public static <T> FloatPriorityQueue<T> synchronizedPq(int initialCapacity) {
		return new FloatPriorityQueue<T>(Queues.synchronizedQueue(new PriorityQueue<>(Math.max(1, initialCapacity))));
	}

	@Override
	public int size() {
		return internalQueue.size();
	}

	@Override
	public boolean isEmpty() {
		return internalQueue.isEmpty();
	}

	@Override
	public boolean contains(Object o) {
		return internalQueue.contains(ScoredValue.of(0, o));
	}

	@NotNull
	@Override
	public Iterator<T> iterator() {
		var it = internalQueue.iterator();
		return new Iterator<T>() {
			@Override
			public boolean hasNext() {
				return it.hasNext();
			}

			@Override
			public T next() {
				return getValueOrNull(it.next());
			}
		};
	}

	private T getValueOrNull(ScoredValue<T> scoredValue) {
		if (scoredValue == null) {
			return null;
		} else {
			return scoredValue.getValue();
		}
	}

	@NotNull
	@Override
	public Object[] toArray() {
		return internalQueue.stream().map(this::getValueOrNull).toArray(Object[]::new);
	}

	@NotNull
	@Override
	public <T1> T1[] toArray(@NotNull T1[] a) {
		return internalQueue
				.stream()
				.map(this::getValueOrNull)
				.toArray(i -> (T1[]) Array.newInstance(a.getClass().getComponentType(), i));
	}

	@Deprecated
	@Override
	public boolean add(T t) {
		return internalQueue.add(ScoredValue.of(0, t));
	}

	public boolean addTop(T t) {
		return internalQueue.add(ScoredValue.of(Integer.MAX_VALUE, t));
	}

	public boolean add(T t, float score) {
		return internalQueue.add(ScoredValue.of(score, t));
	}

	@Override
	public boolean remove(Object o) {
		return internalQueue.remove(ScoredValue.of(0, o));
	}

	@Override
	public boolean containsAll(@NotNull Collection<?> c) {
		return internalQueue.containsAll(c.stream().map(val -> ScoredValue.of(0, val)).collect(Collectors.toList()));
	}

	@Override
	public boolean addAll(@NotNull Collection<? extends T> c) {
		return internalQueue.addAll(c.stream().map(val -> ScoredValue.of(0, (T) val)).collect(Collectors.toList()));
	}

	@Override
	public boolean removeAll(@NotNull Collection<?> c) {
		return internalQueue.removeAll(c.stream().map(val -> ScoredValue.of(0, val)).collect(Collectors.toList()));
	}

	@Override
	public boolean retainAll(@NotNull Collection<?> c) {
		return internalQueue.retainAll(c.stream().map(val -> ScoredValue.of(0, val)).collect(Collectors.toList()));
	}

	@Override
	public void clear() {
		internalQueue.clear();
	}

	@Override
	public boolean offer(T t) {
		return offer(ScoredValue.of(0, t));
	}

	public boolean offer(T t, float score) {
		return offer(ScoredValue.of(score, t));
	}

	public boolean offer(ScoredValue<T> value) {
		return this.internalQueue.offer(value);
	}

	@Override
	public T remove() {
		return getValueOrNull(internalQueue.remove());
	}

	@Override
	public T poll() {
		return getValueOrNull(internalQueue.poll());
	}

	@Override
	public T element() {
		return getValueOrNull(internalQueue.element());
	}

	@Override
	public T peek() {
		return getValueOrNull(internalQueue.peek());
	}

	public void forEachItem(Consumer<ScoredValue<T>> action) {
		internalQueue.forEach(action);
	}

	public Stream<ScoredValue<T>> streamItems() {
		return internalQueue.stream();
	}
}
